"""DBF reader.

Copyright 2024 Edgar Ramírez-Mondragón.
"""

from __future__ import annotations

import collections
import datetime
import typing as t
from pathlib import Path

from dbfread import DBF
from dbfread.dbf import expand_year, ifind
from dbfread.exceptions import DBFNotFound
from dbfread.field_parser import FieldParser

if t.TYPE_CHECKING:
    from fsspec import AbstractFileSystem


class FilesystemDBF(DBF):  # type: ignore[misc]
    """A DBF implementation that can open files in arbitrary filesystems."""

    def __init__(  # noqa: PLR0913
        self,
        filename: str,
        filesystem: AbstractFileSystem,
        *,
        encoding: str | None = None,
        ignorecase: bool = True,
        lowernames: bool = False,
        parserclass: type[FieldParser] = FieldParser,
        recfactory: type[dict[str, t.Any]] = collections.OrderedDict,
        load: bool = False,
        raw: bool = False,
        ignore_missing_memofile: bool = False,
        char_decode_errors: str = "strict",
    ) -> None:
        """Create a new DBF reader.

        Raises:
            DBFNotFound: If the dBase file is not found.
        """
        self.filesystem = filesystem
        self.encoding = encoding
        self.ignorecase = ignorecase
        self.lowernames = lowernames
        self.parserclass = parserclass
        self.raw = raw
        self.ignore_missing_memofile = ignore_missing_memofile
        self.char_decode_errors = char_decode_errors

        self.date: datetime.date | None

        self.recfactory = recfactory

        # Name part before .dbf is the table name
        self.name = Path(filename).stem
        self._records = None
        self._deleted = None

        if ignorecase:
            self.filename = ifind(filename)
            if not self.filename:
                msg = f"could not find file {filename!r}"
                raise DBFNotFound(msg)
        else:
            self.filename = filename

        # Filled in by self._read_headers()
        self.memofilename = None
        self.header: t.Any | None = None
        self.fields: list[t.Any] = []  # namedtuples
        self.field_names: list[str] = []  # strings

        with self.filesystem.open(self.filename, mode="rb") as infile:
            self._read_header(infile)
            self._read_field_headers(infile)
            self._check_headers()

            try:
                self.date = datetime.date(
                    expand_year(self.header.year),  # type: ignore[attr-defined]
                    self.header.month,  # type: ignore[attr-defined]
                    self.header.day,  # type: ignore[attr-defined]
                )
            except ValueError:
                # Invalid date or '\x00\x00\x00'.
                self.date = None

        self.memofilename = self._get_memofilename()

        if load:
            self.load()

    def _count_records(self, record_type: bytes = b" ") -> int:
        """Count records in the table.

        Args:
            record_type: The record type to count.

        Returns:
            The number of records.
        """
        count = 0

        with self.filesystem.open(self.filename, "rb") as infile:
            # Skip to first record.
            infile.seek(self.header.headerlen, 0)  # type: ignore[union-attr]

            while True:
                sep = infile.read(1)
                if sep == record_type:
                    count += 1
                    self._skip_record(infile)
                elif sep in {b"\x1a", b""}:
                    # End of records.
                    break
                else:
                    self._skip_record(infile)

        return count

    def _iter_records(
        self,
        record_type: bytes = b" ",
    ) -> t.Generator[dict[str, t.Any], None, None]:
        """Iterate over records in the table.

        Args:
            record_type: The record type to iterate over.

        Yields:
            A record.
        """
        with (
            self.filesystem.open(
                self.filename,
                "rb",
            ) as infile,
            self._open_memofile() as memofile,
        ):
            # Skip to first record.
            infile.seek(self.header.headerlen, 0)  # type: ignore[union-attr]

            if not self.raw:
                field_parser = self.parserclass(self, memofile)
                parse = field_parser.parse

            # Shortcuts for speed.
            skip_record = self._skip_record
            read = infile.read

            while True:
                sep = read(1)

                if sep == record_type:
                    if self.raw:
                        items = [
                            (field.name, read(field.length)) for field in self.fields
                        ]
                    else:
                        items = [
                            (field.name, parse(field, read(field.length)))
                            for field in self.fields
                        ]

                    yield self.recfactory(items)

                elif sep in {b"\x1a", b""}:
                    # End of records.
                    break
                else:
                    skip_record(infile)



"""Test tap-dbf.

Copyright 2024 Edgar Ramírez-Mondragón.
"""

from __future__ import annotations

from singer_sdk.testing import get_tap_test_class

from tap_dbf.tap import TapDBF

TestTapDBF = get_tap_test_class(
    TapDBF,
    config={
        "path": "/files/*.dbf",
        "fs_root": "file://tests/data",
        "ignore_missing_memofile": True,
    },
)

TestTapDBFNoRoot = get_tap_test_class(
    TapDBF,
    config={
        "path": "tests/data/files/*.dbf",
        "ignore_missing_memofile": True,
    },
)



"""Singer SDK classes.

Copyright 2024 Edgar Ramírez-Mondragón.
"""

from __future__ import annotations

import builtins
import logging
import typing as t
from pathlib import Path
from urllib.parse import parse_qsl, urlparse, urlunparse

import fsspec
import singer_sdk.typing as th
from singer_sdk import Stream, Tap

from tap_dbf.client import FilesystemDBF

if t.TYPE_CHECKING:
    import sys
    from os import PathLike
    from types import TracebackType

    from dbfread.dbf import DBFField
    from fsspec import AbstractFileSystem
    from singer_sdk.helpers.types import Context

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    OpenFunc = t.Callable[[PathLike[bytes], str], t.BinaryIO]
    RawRecord = dict[str, t.Any]

logger = logging.getLogger(__name__)


def dbf_field_to_jsonschema(field: DBFField) -> dict[str, t.Any]:
    """Map a .dbf data type to a JSON schema.

    Args:
        field: The field to map.

    Returns:
        A JSON schema.
    """
    d: dict[str, t.Any] = {"type": ["null"]}
    if field.type == "N":
        if field.decimal_count == 0:
            d["type"].append("integer")
        else:
            d["type"].append("number")
    elif field.type in {"+", "I"}:
        d["type"].append("integer")
    elif field.type in {"B", "F", "O", "Y"}:
        d["type"].append("number")
    elif field.type == "L":
        d["type"].append("boolean")
    elif field.type == "D":
        d["type"].append("string")
        d["format"] = "date-time"
    elif field.type in {"@", "T"}:
        d["type"].append("string")
        d["format"] = "time"
    else:
        d["type"].append("string")
        d["maxLength"] = field.length

    return d


class PatchOpen:
    """Context helper to patch the builtin open function."""

    def __init__(self: PatchOpen, fs: AbstractFileSystem) -> None:
        """Patch builtins.open with a custom open function.

        Args:
            fs: The filesystem instance to use.
        """
        self.old_impl = _patch_open(fs.open)

    def __enter__(self: Self) -> Self:
        """Create a context for the patched function.

        Returns:
            The PatchOpen context.
        """
        return self

    def __exit__(
        self: PatchOpen,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit context and revert patch.

        Args:
            exc_type: The exception type.
            exc_val: The exception value.
            exc_tb: The exception traceback.
        """
        _patch_open(self.old_impl)


def _patch_open(func: OpenFunc) -> OpenFunc:
    """Patch `builtins.open` with `func`.

    Args:
        func: The function to patch `builtins.open` with.

    Returns:
        The original `builtins.open` function.
    """
    old_impl = builtins.open
    builtins.open = func  # type: ignore[assignment]
    return old_impl  # type: ignore[return-value]


class DBFStream(Stream):
    """A dBase file stream."""

    def __init__(
        self: DBFStream,
        tap: Tap,
        filepath: str,
        filesystem: AbstractFileSystem,
        *,
        ignore_missing_memofile: bool = False,
    ) -> None:
        """Create a new .DBF file stream.

        Args:
            tap: The tap instance.
            filepath: The path to the .DBF file.
            filesystem: The filesystem instance to use.
            ignore_missing_memofile: Whether to ignore missing .DBT files.
        """
        self.filepath = filepath
        self.filesystem = filesystem
        name = Path(filepath).stem

        self._table: FilesystemDBF = FilesystemDBF(
            filepath,
            ignorecase=False,
            ignore_missing_memofile=ignore_missing_memofile,
            filesystem=filesystem,
        )

        self._fields = list(self._table.fields)
        schema: dict[str, t.Any] = {"properties": {}}
        self.primary_keys = []

        for field in self._fields:
            schema["properties"][field.name] = dbf_field_to_jsonschema(field)
            if field.type == "+":
                self.primary_keys.append(field.name)

        schema["properties"]["_sdc_filepath"] = {"type": ["string"]}
        schema["properties"]["_sdc_row_index"] = {"type": ["integer"]}

        super().__init__(tap, schema=schema, name=name)

    def get_records(
        self: DBFStream,
        _: Context | None = None,
    ) -> t.Iterable[RawRecord]:
        """Get .DBF rows.

        Yields:
            A row of data.
        """
        for index, row in enumerate(self._table):
            row["_sdc_filepath"] = self.filepath
            row["_sdc_row_index"] = index
            yield row


class TapDBF(Tap):
    """A singer tap for .DBF files."""

    name = "tap-dbf"
    config_jsonschema = th.PropertiesList(
        th.Property(
            "path",
            th.StringType,
            required=True,
            description=(
                "Glob expression where the files are located. Stream names will be "
                "extracted from the file name."
            ),
        ),
        th.Property(
            "fs_root",
            th.StringType,
            default="file://",
            description="The root of the filesystem to read from.",
        ),
        th.Property(
            "ignore_missing_memofile",
            th.BooleanType,
            default=False,
            description=(
                "Whether to proceed reading the file even if the [memofile] is not "
                "present."
            ),
        ),
        th.Property(
            "s3",
            th.ObjectType(
                th.Property(
                    "key",
                    th.StringType,
                    secret=True,
                    description="The AWS key ID.",
                ),
                th.Property(
                    "secret",
                    th.StringType,
                    secret=True,
                    description="The AWS secret key.",
                ),
                th.Property(
                    "endpoint_url",
                    th.StringType,
                    description="The S3 endpoint URL.",
                    examples=[
                        "https://localhost:9000",
                    ],
                ),
            ),
            description="S3 configuration.",
        ),
        th.Property(
            "gcs",
            th.ObjectType(
                th.Property(
                    "token",
                    th.StringType,
                    description="OAuth 2.0 token for GCS.",
                ),
            ),
            description="GCS configuration.",
        ),
    ).to_dict()

    def discover_streams(self: TapDBF) -> list[DBFStream]:
        """Discover .DBF files in the path.

        Returns:
            A list of streams.
        """
        streams = []

        fs_root: str = self.config["fs_root"]
        url = urlparse(fs_root)
        protocol = url.scheme

        storage_options = {
            **dict(parse_qsl(url.query)),
            **self.config.get(protocol, {}),
        }

        fs: AbstractFileSystem = fsspec.filesystem(url.scheme, **storage_options)

        path = url.path + self.config["path"]
        if not url.hostname:
            hostname, path = path.split("/", 1)
        else:
            hostname = url.hostname

        full_path = urlunparse(url._replace(query="", netloc=hostname, path=path))

        for match in fs.glob(full_path):
            self.logger.info("Extracting from %s://%s", url.scheme, match)
            stream = DBFStream(
                tap=self,
                filepath=match,
                filesystem=fs,
                ignore_missing_memofile=self.config["ignore_missing_memofile"],
            )

            streams.append(stream)

        return streams
